from Tkinter import *

import random
import time

def generate_missile_damage(damage):
    template = []

    n = 1
    while True:
        template.append(0)

        ns = range(n, 0, -1)
        template += map(lambda i: -i, ns)
        template += ns

        if len(template) > damage:
            return template

        n += 1

def generate_laser_damage(damage):
    template = [0,0,0]

    n = 1
    while True:
        ns = range(n, 0, -1)
        next_section = map(lambda i: -i, ns) + ns + [0]

        template += next_section
        template += next_section
        template += next_section

        if len(template) > damage:
            return template

        n += 1

armor_height = 10
armor_width = 10

class ArmorPierce:
    def __init__(self):
        self.parent = Tk()
        self.parent.title('Aurora Armor Calculator:')
        self.parent.geometry("{}x{}+0+0".format(950, 750))

        self.armor = []
        self.max_shield_strength = 0
        self.shield_strength = 0
        self.shield_recharge_time = 0

        self.weapon_hits_list = []
        self.weapon_volley_intervals = []

        self.setup_gui()

        self.parent.mainloop()

    def setup_gui(self):
        self.parent.columnconfigure(0, weight=1)

        self.canvas = Canvas(self.parent, bd=1, relief=RIDGE, width=800, height=750)
        self.canvas.grid(row=0, column=0, rowspan=400, sticky=N+S+E+W)

        self.armor_label = Label(self.parent, text='Armor (Layers-Width, e.g. 7-32):')
        self.armor_label.grid(row=0, column=1, columnspan=2, sticky=W)

        self.armor_var = StringVar()
        self.armor_var.trace("w", lambda name, index, mode: self.redo_armor())
        self.armor_entry = Entry(self.parent, textvariable=self.armor_var)
        self.armor_var.set('0-0')
        self.armor_entry.grid(row=1, column=1, columnspan=2, sticky=W)

        self.shield_label = Label(self.parent, text='Shield (Strength-Recharge Time, e.g. 20-300):')
        self.shield_label.grid(row=2, column=1, columnspan=2, sticky=W)

        self.shield_var = StringVar()
        self.shield_var.set('0-0')
        self.shield_var.trace("w", lambda name, index, mode: self.redo_shield())
        self.shield_entry = Entry(self.parent, textvariable=self.shield_var)
        self.shield_entry.grid(row=3, column=1, columnspan=2, sticky=W)

        self.weapon_type_label = Label(self.parent, text='Weapon Type:')
        self.weapon_type_label.grid(row=4, column=1, sticky=W)

        self.weapon_type_var = StringVar()
        self.weapon_type = OptionMenu(self.parent, self.weapon_type_var, 'Missile', 'Laser')
        self.weapon_type.grid(row=5, column=1, sticky=W)

        self.weapon_damage_label = Label(self.parent, text='Weapon Damage:')
        self.weapon_damage_label.grid(row=6, column=1, sticky=W)

        self.weapon_damage_var = StringVar()
        self.weapon_damage_var.set('0')
        self.weapon_damage_entry = Entry(self.parent, textvariable=self.weapon_damage_var)
        self.weapon_damage_entry.grid(row=6, column=2, sticky=W)

        self.shots_label = Label(self.parent, text='Shots in Volley:')
        self.shots_label.grid(row=7, column=1, sticky=W)

        self.shots_count_var = StringVar()
        self.shots_count_var.set('0')
        self.shots_count = Entry(self.parent, textvariable=self.shots_count_var)
        self.shots_count.grid(row=7, column=2, sticky=W)

        self.volley_interval_label = Label(self.parent, text='Volley Interval:')
        self.volley_interval_label.grid(row=8, column=1, sticky=W)

        self.volley_interval_var = StringVar()
        self.volley_interval_var.set('0')
        self.volley_interval_entry = Entry(self.parent, textvariable=self.volley_interval_var)
        self.volley_interval_entry.grid(row=8, column=2, sticky=W)

        self.hit_chance_label = Label(self.parent, text='Hit Chance (%):')
        self.hit_chance_label.grid(row=9, column=1, sticky=W)

        self.hit_chance_var = StringVar()
        self.hit_chance_var.set('0')
        self.hit_chance_entry = Entry(self.parent, textvariable=self.hit_chance_var)
        self.hit_chance_entry.grid(row=9, column=2, sticky=W)

        self.add_weapon_hit_button = Button(self.parent, text='Add', command=self.add_hit)
        self.add_weapon_hit_button.grid(row=10, column=1, sticky=W)

        self.weapon_hits = Listbox(self.parent)
        self.weapon_hits.grid(row=11, rowspan=2, column=1, columnspan=2, sticky=N+S+E+W)

        self.reset_button = Button(self.parent, text='Reset', command=self.reset)
        self.reset_button.grid(row=13, column=1, sticky=W)

        self.remove_last_button = Button(self.parent, text='Clear', command=self.clear)
        self.remove_last_button.grid(row=13, column=2, sticky=W)

        self.calc_armor_pierce_label = Label(self.parent, text='Calculate Min Hits:')
        self.calc_armor_pierce_label.grid(row=14, column=1, sticky=W)
        self.calc_armor_pierce_label.bind("<Button-1>", lambda _: self.calc_armor_pierce.invoke())

        self.calc_armor_pierce_var = BooleanVar()
        self.calc_armor_pierce_var.set(False)
        self.calc_armor_pierce = Checkbutton(self.parent, variable=self.calc_armor_pierce_var)
        self.calc_armor_pierce.grid(row=14, column=2, sticky=W)

        self.damage_target_label = Label(self.parent, text='Damage Target:')
        self.damage_target_label.grid(row=15, column=1, sticky=W)

        self.damage_target_var = StringVar()
        self.damage_target_var.set('0')
        self.damage_target_entry = Entry(self.parent, textvariable=self.damage_target_var)
        self.damage_target_entry.grid(row=15, column=2, sticky=W)

        self.calculate_hits_button = Button(self.parent, text='Calculate', command=self.calc_button_press)
        self.calculate_hits_button.grid(row=16, column=1, sticky=W)

        self.result_count = 0
        self.result_label = Label(self.parent, text='Pierced Damage: 0')
        self.result_label.grid(row=17, column=1, columnspan=2, sticky=W)

        self.time = 0
        self.time_label = Label(self.parent, text='Time Elapsed (seconds): 0')
        self.time_label.grid(row=18, column=1, columnspan=2, sticky=W)

        self.redos_var = StringVar()
        self.redos_var.set('0')
        self.redos_entry = Entry(self.parent, textvariable=self.redos_var)
        self.redos_entry.grid(row=19, column=1, columnspan=2, sticky=W)

        self.redo_button = Button(self.parent, text='Redo Calculation', command=self.redo)
        self.redo_button.grid(row=20, column=1, columnspan=2, sticky=W)

        self.average_result_count = 0
        self.average_result_label = Label(self.parent, text='Average Pierced Damage: 0')
        self.average_result_label.grid(row=21, column=1, columnspan=2, sticky=W)

        self.average_time = 0
        self.average_time_label = Label(self.parent, text='Average Time Elapsed (seconds): 0')
        self.average_time_label.grid(row=22, column=1, columnspan=2, sticky=W)

    def toggle_calc_armor_pierce(self):
        return

    def clear(self):
        self.weapon_hits_list = []
        self.weapon_volley_intervals = []
        self.weapon_hits.delete(0, END)

    def add_hit(self):
        #Add important information to a list (weapon type, weapon damage, number of hits)
        weapon_type = self.weapon_type_var.get()
        weapon_damage = int(self.weapon_damage_var.get())
        shots = int(self.shots_count_var.get())
        volley_interval = int(self.volley_interval_var.get())
        hit_chance = float(self.hit_chance_var.get())

        self.weapon_hits_list.append((weapon_type, volley_interval, weapon_damage, shots, hit_chance))
        self.weapon_volley_intervals.append(0) #0 Means redo to fire, every weapon starts ready to fire
        self.weapon_hits.insert(END, '{} ({}s): {} x{} ({}%)'.format(weapon_type, volley_interval, weapon_damage, shots, hit_chance))

    def set_result_count(self, amount):
        self.result_count = amount

        if not self.calc_armor_pierce_var.get():
            self.result_label['text'] = 'Pierced Damage: {}'.format(self.result_count)
        else:
            self.result_label['text'] = 'Hits to Damage: {}'.format(self.result_count)

    def increment_result_count(self):
        self.set_result_count(self.result_count + 1)

    def set_average_result_count(self, amount):
        self.average_result_count = amount

        if not self.calc_armor_pierce_var.get():
            self.average_result_label['text'] = 'Average Pierced Damage: {}'.format(self.average_result_count)
        else:
            self.average_result_label['text'] = 'Average Hits to Damage: {}'.format(self.average_result_count)

    def increment_time_by(self, amount):
        self.set_time(self.time + amount)

    def set_time(self, value):
        self.time = value

        self.time_label['text'] = 'Time Elapsed (seconds): {}'.format(self.time)

    def set_average_time(self, value):
        self.average_time = value

        self.average_time_label['text'] = 'Average Time Elapsed (seconds): {}'.format(self.average_time)

    # Do damage to the armor according to the damage profile.
    def do_damage(self, damage_profile, damage):
        if damage > self.shield_strength:
            damage = int(damage - self.shield_strength)
            self.shield_strength = 0
        else:
            self.shield_strength -= damage
            return

        hit_index = random.randint(0, len(self.armor))

        pierced_armor = False

        for i in xrange(damage):
            if i > len(damage_profile):
                break
            real_index = hit_index + damage_profile[i]
            if real_index < 0:
                real_index = len(self.armor) + real_index
            elif real_index >= len(self.armor):
                real_index -= len(self.armor)

            hit_armor = False
            for r, damaged in enumerate(self.armor[real_index]):
                if not damaged:
                    hit_armor = True
                    self.armor[real_index][r] = True
                    break

            if not hit_armor:
                self.increment_result_count()

        self.draw_armor()

    def advance_time(self, amount):
        self.increment_time_by(amount)

        for i, v in enumerate(self.weapon_volley_intervals):
            self.weapon_volley_intervals[i] = v - amount

        self.recharge_shield(amount)

    def advance_step(self):
        self.advance_time(5)

    def reset_volley_intervals(self):
        for i in xrange(len(self.weapon_volley_intervals)):
            self.weapon_volley_intervals[i] = 0

    def calc_button_press(self):
        if not self.calc_armor_pierce_var.get():
            self.reset_volley_intervals()

            self.calchit()
        else:
            self.calchits()

    # If not advanced, just calculate one hit from the weapon stats listed.
    # If advanced, then calculate one round of hits from all the weapons listed in weapon_hits_list.
    def calchit(self):
        self.advance_step()

        weapon_hits = 0
        for i, (weapon_type, volley_interval, weapon_damage, shots, hit_chance) in enumerate(self.weapon_hits_list):
            if self.weapon_volley_intervals[i] <= 0:
                for s in xrange(shots):
                    if random.random() < (hit_chance / 100.0):
                        if weapon_type == 'Missile':
                            self.do_damage(generate_missile_damage(weapon_damage), weapon_damage)
                        elif weapon_type == 'Laser':
                            self.do_damage(generate_laser_damage(weapon_damage), weapon_damage)

                        weapon_hits += 1

                        if self.calc_armor_pierce_var.get() and self.result_count > int(self.damage_target_var.get()):
                            return weapon_hits

                self.weapon_volley_intervals[i] = volley_interval

        return weapon_hits

    def calchits(self):
        if not self.calc_armor_pierce_var.get():
            return self.calchit()
        else:
            i = 0
            while self.result_count <= int(self.damage_target_var.get()):
                i += self.calchit()

            self.set_result_count(i)

            return i

    def reset(self):
        self.redo_armor()
        self.redo_shield()

        self.reset_volley_intervals()

        self.set_result_count(0)
        self.set_average_result_count(0)
        self.set_time(0)
        self.set_average_time(0)

    def redo(self):
        total = 0.0
        total_time = 0.0
        for i in xrange(int(self.redos_var.get())):
            self.reset()
            total += self.calchits()
            total_time += self.time

            if i > 0:
                self.parent.wm_title('{}% complete ({}, {} s)'.format(float(i + 1) / float(self.redos_var.get()) * 100, total / float(i), total_time / float(i)))

        self.set_average_time(total_time / float(self.redos_var.get()))
        self.set_average_result_count(total / float(self.redos_var.get()))

    def draw_armor(self):
        self.canvas.delete('all')

        for x, row in enumerate(self.armor):
            for y, piece in enumerate(row):
                if piece:
                    self.canvas.create_rectangle(x * armor_width + 5, y * armor_height + 5, x * armor_width + armor_width + 5, y * armor_height + armor_height + 5, outline='red', fill='red')
                else:
                    self.canvas.create_rectangle(x * armor_width + 5, y * armor_height + 5, x * armor_width + armor_width + 5, y * armor_height + armor_height + 5)

    def recharge_shield(self, time_elapsed):
        if self.max_shield_strength <= 0 or self.shield_recharge_time <= 0:
            return
        self.shield_strength += time_elapsed * float(self.max_shield_strength) / float(self.shield_recharge_time)

        self.shield_strength = min(self.shield_strength, self.max_shield_strength)

    def redo_shield(self):
        try:
            self.max_shield_strength, self.shield_recharge_time = map(float, self.shield_var.get().split('-'))

            self.shield_strength = self.max_shield_strength
        except:
            self.shield_strength = 0
            self.max_shield_strength = 0
            self.shield_recharge_time = 0

    def redo_armor(self):
        try:
            width, height = map(int, self.armor_var.get().split('-'))

            self.armor = []

            for x in xrange(height):
                self.armor.append([])
                for y in xrange(width):
                    self.armor[-1].append(False)

            self.draw_armor()
        except:
            self.armor = []

if __name__ == '__main__':
    main = ArmorPierce()
